/*
 * linux/include/asm-arm/arch-nexuspci/dma.h
 *
 * Architecture DMA routines
 *
 * Copyright (C) 1998 Philip Blundell
 */

#ifndef __ASM_ARCH_DMA_H
#define __ASM_ARCH_DMA_H

/* NexusPCI has no DMA */

#warning No DMA on this platform

#endif /* _ASM_ARCH_DMA_H */

